//----------------------------------------------------------------------------
//
// File:        ti994a.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "compress.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "tms9919.hpp"
#include "cartridge.hpp"
#include "device.hpp"
#include "ti994a.hpp"

#define RAM_BANK_SIZE	0x1000
#define GROM_BANK_SIZE	0x2000

extern "C" UCHAR  CpuMemory [ 0x10000 ];

extern "C" {
    void *CRU_Object;
};

#if defined ( __GNUC__ )
void swab ( char *src, char *dest, unsigned length )
{
    for ( unsigned i = 0; i < length; i += 2 ) {
        char lo = src [i];
        char hi = src [i+1];
        dest [i] = hi;
        dest [i+1] = lo;
    }
}
#endif

cTI994A::cTI994A ( cCartridge *_console, cTMS9918A *_vdp, cTMS9919 *_sound )
{
    FUNCTION_ENTRY ( this, "cTI994A ctor", true );

    // Define CRU_Object for _OpCodes.asm
    CRU_Object = this;

    m_CpuMemory = CpuMemory;

    m_GromMemory  = new UCHAR [ 0x10000 ];

    memset ( CpuMemory, 0, 0x10000 );
    memset ( m_GromMemory, 0, 0x10000 );

    memset ( m_CpuMemoryInfo, 0, sizeof ( m_CpuMemoryInfo ));
    memset ( m_GromMemoryInfo, 0, sizeof ( m_GromMemoryInfo ));

    m_CPU = new cTMS9900 ();

    m_VDP = _vdp ? _vdp : new cTMS9918A ();
    m_VDP->SetHost ( this );

    m_VideoMemory = m_VDP->GetMemory ();

    m_SoundGenerator = _sound ? _sound : new cTMS9919;

    // Start off with all memory set as ROM - simulates no memory present
    m_CPU->SetMemory ( MEM_ROM, 0x0000, 0x10000 );

    // No Devices by default - derived classes can add Devices
    m_ActiveCRU = 0;
    memset ( m_Device, 0, sizeof ( m_Device ));

    m_Console   = NULL;
    m_Cartridge = NULL;

    InsertCartridge ( _console );

    m_Cartridge = NULL;
    m_Console   = _console;

    m_VDP_Retrace = 0;

    m_GromPtr             = m_GromMemory;
    m_GromAddress         = 0;
    m_GromLastInstruction = 0;
    m_GromReadShift       = 8;
    m_GromWriteShift      = 8;
    m_GromCounter         = 0;

    UCHAR index;
    m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_BANK_SWITCH );

    index = m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_SCRATCH_PAD );
    for ( USHORT address = 0x8000; address < 0x8300; address += ( USHORT ) 1 ) {
        m_CPU->SetBreakpoint ( address, ( UCHAR ) MEMFLG_ACCESS, false, index );
    }

    index = m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_SOUND );
    m_CPU->SetBreakpoint ( 0x8400, MEMFLG_WRITE, true, index );		// Sound chip Port

    index = m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_VIDEO );
    for ( USHORT address = 0x8800; address < 0x8C00; address += ( USHORT ) 4 ) {
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0000 ), MEMFLG_READ, true, index );	// VDP Read Byte Port
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0002 ), MEMFLG_READ, true, index );	// VDP Read Status Port
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0400 ), MEMFLG_WRITE, false, index );	// VDP Write Byte Port
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0402 ), MEMFLG_WRITE, false, index );	// VDP Write (set) Address Port
    }

    m_CPU->SetMemory ( MEM_ROM, 0x9000, RAM_BANK_SIZE );
    index = m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_SPEECH );
    m_CPU->SetBreakpoint ( 0x9000, MEMFLG_READ, true, index );		// Speech Read Port
    m_CPU->SetBreakpoint ( 0x9400, MEMFLG_WRITE, true, index );		// Speech Write Port

    index  = m_CPU->RegisterBreakpoint ( TrapFunction, this, TRAP_GROM );
    for ( USHORT address = 0x9800; address < 0x9C00; address += ( USHORT ) 2 ) {
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0000 ), ( UCHAR ) MEMFLG_READ, true, index );
        m_CPU->SetBreakpoint (( USHORT ) ( address | 0x0400 ), ( UCHAR ) MEMFLG_WRITE, true, index );
    }
}

cTI994A::~cTI994A ()
{
    FUNCTION_ENTRY ( this, "cTI994A dtor", true );

    for ( unsigned i = 0; i < SIZE ( m_Device ); i++ ) {
        if ( m_Device [i] != NULL ) delete m_Device [i];
    }

    delete [] m_GromMemory;

    delete m_SoundGenerator;
    delete m_CPU;
    delete m_VDP;
    delete m_Console;
}

// Convert data stored x86 style (Little-Endian) to TMS9900 style (Big-Endian)
void LoadMemory ( USHORT *dest, USHORT *src, unsigned length )
{
    FUNCTION_ENTRY ( NULL, "LoadMemory", true );

    swab (( char * ) src, ( char * ) dest, length );
}

cDevice *cTI994A::GetDevice ( ADDRESS address )
{
    FUNCTION_ENTRY ( this, "cTI994A::GetDevice", true );

    return m_Device [ ( address >> 8 ) & 0x0F ];
}

USHORT cTI994A::TrapFunction ( void *ptr, int type, bool read, const ADDRESS address, USHORT value )
{
    cTI994A *pThis = ( cTI994A * ) ptr;
    USHORT retVal = value;

    if ( read == true ) {
        switch ( type ) {
            case TRAP_BANK_SWITCH :
                retVal = pThis->BankSwitch ( address, value );
                break;
            case TRAP_SCRATCH_PAD :
                retVal = pThis->ScratchPadRead ( address & 0x00FF, value );
                break;
            case TRAP_SOUND :
                retVal = pThis->SoundBreakPoint ( address, value );
                break;
            case TRAP_SPEECH :
                retVal = pThis->SpeechBreakPoint ( address, value );
                break;
            case TRAP_VIDEO :
                retVal = pThis->VideoReadBreakPoint ( address & 0xFF02, value );
                break;
            case TRAP_GROM :
                retVal = pThis->GromReadBreakPoint ( address & 0xFF02, value );
                break;
            default :
                fprintf ( stderr, "Invalid index %d for read access in TrapFunction\n", type );
                break;
        };
    } else {
        switch ( type ) {
            case TRAP_BANK_SWITCH :
                retVal = pThis->BankSwitch ( address, value );
                break;
            case TRAP_SCRATCH_PAD :
                retVal = pThis->ScratchPadWrite ( address & 0x00FF, value );
                break;
            case TRAP_SOUND :
                retVal = pThis->SoundBreakPoint ( address, value );
                break;
            case TRAP_SPEECH :
                retVal = pThis->SpeechBreakPoint ( address, value );
                break;
            case TRAP_VIDEO :
                retVal = pThis->VideoWriteBreakPoint ( address & 0xFF02, value );
                break;
            case TRAP_GROM :
                retVal = pThis->GromWriteBreakPoint ( address & 0xFF02, value );
                break;
            default :
                fprintf ( stderr, "Invalid index %d for write access in TrapFunction\n", type );
                break;
        };
    }

    return retVal;
}

USHORT cTI994A::BankSwitch ( const ADDRESS address, USHORT )
{
    FUNCTION_ENTRY ( this, "cTI994A::BankSwitch", true );

    sMemoryRegion *region = m_CpuMemoryInfo [( address >> 12 ) & 0xFE ];
    int newBank = ( address >> 1 ) % region->noBanks;
    region->curBank = region->Bank[newBank];
    ADDRESS baseAddress = ( ADDRESS ) ( address & 0xE000 );
    LoadMemory (( USHORT * ) &CpuMemory [ baseAddress ], ( USHORT * ) region->curBank, RAM_BANK_SIZE );
    region++;
    region->curBank = region->Bank[newBank];
    LoadMemory (( USHORT * ) &CpuMemory [ baseAddress + RAM_BANK_SIZE ], ( USHORT * ) region->curBank, RAM_BANK_SIZE );
    return CpuMemory [ address ];
}

USHORT cTI994A::ScratchPadRead ( const ADDRESS address, USHORT )
{
    FUNCTION_ENTRY ( this, "cTI994A::ScratchPadRead", true );

    WARNING ( "Reading scratch pad RAM from alias" );

    UCHAR *ptr = &CpuMemory [ 0x8300 + address ];

    return ( USHORT ) (( ptr [0] << 8 ) | ptr [1] );
}

USHORT cTI994A::ScratchPadWrite ( const ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::ScratchPadWrite", true );

    FATAL ( "Unsupported write to scratch pad RAM alias" );

    CpuMemory [ 0x8300 + address ] = ( UCHAR ) data;

    return 0;
}

USHORT cTI994A::SoundBreakPoint ( const ADDRESS, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::SoundBreakPoint", true );

    m_SoundGenerator->WriteData ( data );

    return data;
}

USHORT cTI994A::SpeechBreakPoint ( const ADDRESS, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::SpeechBreakPoint", true );

////    SetBreakpoint ( 0x9000, index );	// Speech Read Port
////    SetBreakpoint ( 0x9400, index );	// Speech Write Port

    return data;
}

USHORT cTI994A::VideoReadBreakPoint ( const ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::VideoReadBreakPoint", true );

    switch ( address ) {
        case 0x8800 :
            data = m_VDP->ReadData ();
            break;
        case 0x8802 :
            data = m_VDP->ReadStatus ();
            break;
    }
    return data;
}

USHORT cTI994A::VideoWriteBreakPoint ( const ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::VideoWriteBreakPoint", true );

    switch ( address ) {
        case 0x8C00 :
            m_VDP->WriteData (( UCHAR ) data );
            break;
        case 0x8C02 :
            m_VDP->WriteAddress (( UCHAR ) data );
            break;
    }
    return data;
 }

USHORT cTI994A::GromReadBreakPoint ( const ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::GromReadBreakPoint", true );

    switch ( address ) {
        case 0x9800 :           // Grom/Gram Read Byte Port
            data = *m_GromPtr;
            m_GromAddress++;
            break;
        case 0x9802 :           // Grom/Gram Read Address Port
            data = ( USHORT ) ((( m_GromAddress + 1 ) >> m_GromReadShift ) & 0x00FF );
            m_GromReadShift = 8 - m_GromReadShift;
            m_GromWriteShift = 8;
            break;
    }
    m_GromPtr = &m_GromMemory [ m_GromAddress ];
    return data;
}

USHORT cTI994A::GromWriteBreakPoint ( const ADDRESS address, USHORT data )
{
    FUNCTION_ENTRY ( this, "cTI994A::GromWriteBreakPoint", true );

    sMemoryRegion *memory;
    switch ( address ) {
        case 0x9C00 :           // Grom/Gram Write Byte Port
            memory = m_GromMemoryInfo [ m_GromAddress >> 13 ];
            if ( memory && ( memory->Type == _RAM_ )) *m_GromPtr = ( UCHAR ) data;
            m_GromAddress++;
            break;
        case 0x9C02 :           // Grom/Gram Write (set) Address Port
            m_GromAddress &= ( ADDRESS ) ( 0xFF00 >> m_GromWriteShift );
            m_GromAddress |= ( ADDRESS ) ( data << m_GromWriteShift );
            m_GromWriteShift = 8 - m_GromWriteShift;
            m_GromReadShift = 8;
            break;
    }
    m_GromPtr = &m_GromMemory [ m_GromAddress ];
    return data;
}

extern "C" void WriteCRU ( cTI994A *ti, ADDRESS address, int count, USHORT value )
{
    FUNCTION_ENTRY ( NULL, "cTI994A::WriteCRU", true );

    while ( count-- ) {
        ti->WriteCRU (( ADDRESS ) ( address++ & 0x1FFF ), value & 1 );
        value >>= 1;
    }
}

extern "C" int ReadCRU ( cTI994A *ti, ADDRESS address, int count )
{
    FUNCTION_ENTRY ( NULL, "cTI994A::ReadCRU", true );

    int value = 0;
    address += ( USHORT ) count;
    while ( count-- ) {
        value <<= 1;
        value |= ti->ReadCRU (( ADDRESS ) ( --address & 0x1FFF ));
    }
    return value;
}

void cTI994A::VerticalRetrace ()
{
    FUNCTION_ENTRY ( this, "cTI994A::VerticalRetrace", true );

    m_VDP_Retrace++;
    m_CPU->SignalInterrupt ( 1 );
}

int cTI994A::ReadCRU ( ADDRESS address )
{
    FUNCTION_ENTRY ( this, "cTI994A::ReadCRU", true );

    address <<= 1;
    cDevice *dev = GetDevice ( address );
    return dev ? dev->ReadCRU (( ADDRESS ) (( address - dev->GetCRU ()) >> 1 )) : 1;
}

void cTI994A::WriteCRU ( ADDRESS address, USHORT val )
{
    FUNCTION_ENTRY ( this, "cTI994A::WriteCRU", true );

    address <<= 1;
    cDevice *dev = GetDevice ( address );
    if ( dev == NULL ) return;
    if ( dev->GetCRU () == address ) {
        cCartridge *ctg = m_Cartridge;
        if ( val == 1 ) {
            m_ActiveCRU = address;
            m_Cartridge = NULL;
            InsertCartridge ( dev->GetROM (), false );
            dev->Activate ();
        } else {
            m_ActiveCRU = 0;
            m_Cartridge = dev->GetROM ();
            dev->DeActivate ();
            RemoveCartridge ( dev->GetROM (), false );
        }
        m_Cartridge = ctg;
    } else {
        dev->WriteCRU (( USHORT ) (( address - dev->GetCRU ()) >> 1 ), val );
    }
}

void cTI994A::Run ()
{
    FUNCTION_ENTRY ( this, "cTI994A::Run", true );

    m_CPU->Run ();
}

bool cTI994A::Step ()
{
    FUNCTION_ENTRY ( this, "cTI994A::Step", true );

    return m_CPU->Step ();
}

void cTI994A::Stop ()
{
    FUNCTION_ENTRY ( this, "cTI994A::Stop", true );

    m_CPU->Stop ();
}

bool cTI994A::IsRunning ()
{
    FUNCTION_ENTRY ( this, "cTI994A::IsRunning", true );

    return m_CPU->IsRunning ();
}

static const char ImageHeader[] = "TI-994/A Memory Image File\n\x1A";

void cTI994A::SaveImage ( char *filename )
{
    FUNCTION_ENTRY ( this, "cTI994A::SaveImage", true );

    FILE *file = fopen ( filename, "wb" );

    fwrite ( ImageHeader, 1, sizeof ( ImageHeader ) - 1, file );

    // Name of inserted Cartridge
    short len;
    if ( m_Cartridge ) {
        len = ( short ) strlen ( m_Cartridge->m_Title );
        fwrite ( &len, sizeof ( len ), 1, file );
        fwrite ( m_Cartridge->m_Title, len, 1, file );
    } else {
        len = 0;
        fwrite ( &len, sizeof ( len ), 1, file );
    }

    m_CPU->SaveImage ( file );
    m_VDP->SaveImage ( file );

    fwrite ( &m_GromAddress, sizeof ( m_GromAddress ), 1, file );
    fwrite ( &m_GromLastInstruction, sizeof ( m_GromLastInstruction ), 1, file );
    fwrite ( &m_GromReadShift, sizeof ( m_GromReadShift ), 1, file );
    fwrite ( &m_GromWriteShift, sizeof ( m_GromWriteShift ), 1, file );
    fwrite ( &m_GromCounter, sizeof ( m_GromCounter ), 1, file );

    int i;
    for ( i = 0; i < 16; i++ ) {
        sMemoryRegion *memory = m_CpuMemoryInfo [ i ];
        if (( memory != NULL ) && ( memory->Type == _RAM_ )) {
            SaveBuffer ( RAM_BANK_SIZE, &CpuMemory [ i * RAM_BANK_SIZE ], file );
        }
    }
    for ( i = 0; i < 8; i++ ) {
        sMemoryRegion *memory = m_GromMemoryInfo [ i ];
        if (( memory != NULL ) && ( memory->Type == _RAM_ )) {
            SaveBuffer ( 0x2000, &m_GromMemory [ i * 0x2000 ], file );
        }
    }

    fwrite ( &m_ActiveCRU, sizeof ( m_ActiveCRU ), 1, file );
    if ( m_ActiveCRU ) {
        cDevice *dev = GetDevice ( m_ActiveCRU );
        dev->SaveImage ( file );
    }

    fclose ( file );
}

char *cTI994A::LoadImage ( char *filename )
{
    FUNCTION_ENTRY ( this, "cTI994A::LoadImage", true );

    FILE *file = fopen ( filename, "rb" );
    if ( file == NULL ) return "<< Unknown >>";

    char buffer [ sizeof ( ImageHeader ) - 1 ];
    fread ( buffer, 1, sizeof ( buffer ), file );

    //   Make sure cartridge(s) match those currently loaded
    short len;
    static char nameBuffer [ 256 ];
    fread ( &len, sizeof ( len ), 1, file );
    int ok = true;
    if ( len ) {
        if ( len > 256 ) len = 256;
        fread ( nameBuffer, len, 1, file );
        if (( m_Cartridge == NULL ) || strncmp ( nameBuffer, m_Cartridge->m_Title, len )) ok = false;
    } else {
        nameBuffer [0] = 0;
        if ( m_Cartridge != NULL ) ok = false;
    }
    if ( ! ok ) {
        fclose ( file );
        return nameBuffer;
    }

    m_CPU->LoadImage ( file );
    m_VDP->LoadImage ( file );

    fread ( &m_GromAddress, sizeof ( m_GromAddress ), 1, file );
    fread ( &m_GromLastInstruction, sizeof ( m_GromLastInstruction ), 1, file );
    fread ( &m_GromReadShift, sizeof ( m_GromReadShift ), 1, file );
    fread ( &m_GromWriteShift, sizeof ( m_GromWriteShift ), 1, file );
    fread ( &m_GromCounter, sizeof ( m_GromCounter ), 1, file );
    m_GromPtr = &m_GromMemory [ m_GromAddress ];

    for ( unsigned i = 0; i < 16; i++ ) {
        sMemoryRegion *memory = m_CpuMemoryInfo [ i ];
        if (( memory != NULL ) && ( memory->Type == _RAM_ )) {
            LoadBuffer ( RAM_BANK_SIZE, &CpuMemory [ i * RAM_BANK_SIZE ], file );
        }
    }
    for ( unsigned i = 0; i < 8; i++ ) {
        sMemoryRegion *memory = m_GromMemoryInfo [ i ];
        if (( memory != NULL ) && ( memory->Type == _RAM_ )) {
            LoadBuffer ( GROM_BANK_SIZE, &m_GromMemory [ i * 0x2000 ], file );
        }
    }

    if ( m_ActiveCRU ) {
        cDevice *dev = GetDevice ( m_ActiveCRU );
        dev->DeActivate ();
        cCartridge *ctg = m_Cartridge;
        m_Cartridge = dev->GetROM ();
        RemoveCartridge ( dev->GetROM (), false );
        m_Cartridge = ctg;
    }
    fread ( &m_ActiveCRU, sizeof ( m_ActiveCRU ), 1, file );
    if ( m_ActiveCRU ) {
        cDevice *dev = GetDevice ( m_ActiveCRU );
        dev->LoadImage ( file );
        cCartridge *ctg = m_Cartridge;
        m_Cartridge = NULL;
        InsertCartridge ( dev->GetROM (), false );
        dev->Activate ();
        m_Cartridge = ctg;
    }

    fclose ( file );

    Refresh ( true );
    return NULL;
}

void cTI994A::Reset ()
{
    FUNCTION_ENTRY ( this, "cTI994A::Reset", true );

    if ( m_CPU ) {
        m_CPU->SetPC ( * ( ADDRESS * ) &CpuMemory [ 0x0002 ] );
        m_CPU->SetWP ( * ( ADDRESS * ) &CpuMemory [ 0x0000 ] );
        m_CPU->SetST ( 0x0000 );
    }
    if ( m_VDP ) m_VDP->Reset ();
}

void cTI994A::AddDevice ( cDevice *dev )
{
    FUNCTION_ENTRY ( this, "cTI994A::AddDevice", true );

    int index = ( dev->GetCRU () & 0x0F00 ) >> 8;
    if ( m_Device [index] != NULL ) return;
    m_Device [index] = dev;
}

void cTI994A::InsertCartridge ( cCartridge *cartridge, bool reset )
{
    FUNCTION_ENTRY ( this, "cTI994A::InsertCartridge", true );

    if ( m_Cartridge != NULL ) {
        cCartridge *ctg = m_Cartridge;
        RemoveCartridge ( ctg );
        delete ctg;
    }
    if ( cartridge == NULL ) return;
    m_Cartridge = cartridge;

    for ( unsigned i = 0; i < SIZE ( m_CpuMemoryInfo ); i++ ) {
        if ( m_Cartridge->CpuMemory [i].noBanks > 0 ) {
            m_CpuMemoryInfo [i] = &m_Cartridge->CpuMemory [i];
            m_CpuMemoryInfo [i]->curBank = m_CpuMemoryInfo [i]->Bank[0];
            if ( m_CpuMemoryInfo [i]->noBanks > 1 ) {
                UCHAR bkp = m_CPU->GetBreakpoint ( TrapFunction, TRAP_BANK_SWITCH );
                USHORT address = ( USHORT ) ( i << 12 );
                for ( unsigned j = 0; j < RAM_BANK_SIZE; j++ ) {
                    m_CPU->SetBreakpoint ( address++, MEMFLG_WRITE, true, bkp );
                }
            }
            eMemoryAccess memType = ( m_CpuMemoryInfo [i]->Type == _RAM_ ) ? MEM_RAM : MEM_ROM;
            m_CPU->SetMemory ( memType, ( ADDRESS ) ( i << 12 ), RAM_BANK_SIZE );
            LoadMemory (( USHORT * ) &CpuMemory [ i << 12 ], ( USHORT * ) m_CpuMemoryInfo [i]->curBank, RAM_BANK_SIZE );
        }
    }

    for ( unsigned i = 0; i < SIZE ( m_GromMemoryInfo ); i++ ) {
        if ( m_Cartridge->GromMemory[i].noBanks > 0 ) {
            m_GromMemoryInfo [i] = &m_Cartridge->GromMemory [i];
            memcpy ( &m_GromMemory [ i << 13 ], m_GromMemoryInfo [i]->curBank, 0x2000 );
        }
    }

    if ( reset ) Reset ();
}

void cTI994A::RemoveCartridge ( cCartridge *cartridge, bool reset )
{
    FUNCTION_ENTRY ( this, "cTI994A::RemoveCartridge", true );

    if ( cartridge != m_Cartridge ) return;

    if ( m_Cartridge != NULL ) {
        for ( unsigned i = 0; i < SIZE ( m_Cartridge->CpuMemory ); i++ ) {
            if ( m_Cartridge->CpuMemory [i].noBanks == 0 ) continue;
//
// If this bank is RAM & Battery backed - update the .ctg file
//
            if ( m_Cartridge->CpuMemory [i].noBanks > 1 ) {
                // Clears bankswitch breakpoint for ALL regions!
                UCHAR bkp = m_CPU->GetBreakpoint ( TrapFunction, TRAP_BANK_SWITCH );
                m_CPU->ClearBreakpoint ( bkp );
            }
        }
        for ( unsigned i = 0; i < SIZE ( m_GromMemoryInfo ); i++ ) {
//          if ( Cartridge->GromMemory [i].noBanks ) && RAM && battery
//          if ( Cartridge->CpuMemory [i].noBanks ) && RAM && battery
//
// If this bank is RAM & Battery backed - update the .ctg file
//
        }
    }
    for ( unsigned i = 0; i < SIZE ( m_Cartridge->CpuMemory ); i++ ) {
        if ( m_Cartridge && ( m_Cartridge->CpuMemory [i].noBanks == 0 )) continue;
        if ( m_Console && m_Console->CpuMemory[i].curBank  ) {
            eMemoryAccess memType = ( m_Console->CpuMemory[i].Type == _RAM_ ) ? MEM_RAM : MEM_ROM;
            m_CPU->SetMemory ( memType, ( ADDRESS ) ( i << 12 ), RAM_BANK_SIZE );
            m_CpuMemoryInfo [i] = &m_Console->CpuMemory [i];
            // Don't clear out memory!?
            LoadMemory (( USHORT * ) &CpuMemory [ i << 12 ], ( USHORT * ) m_CpuMemoryInfo [i]->curBank, RAM_BANK_SIZE );
        } else {
            m_CpuMemoryInfo [i] = NULL;
            m_CPU->SetMemory ( MEM_ROM, ( ADDRESS ) ( i << 12 ), RAM_BANK_SIZE );
            memset (( USHORT * ) &CpuMemory [ i << 12 ], 0, RAM_BANK_SIZE );
        }
    }
    for ( unsigned i = 0; i < SIZE ( m_GromMemoryInfo ); i++ ) {
        if ( m_Cartridge && ( m_Cartridge->GromMemory [i].noBanks == 0 )) continue;
        if ( m_Console && m_Console->GromMemory[i].curBank ) {
            m_GromMemoryInfo [i] = &m_Console->GromMemory [i];
            memcpy (( USHORT * ) &m_GromMemory [ i << 13 ], ( USHORT * ) m_GromMemoryInfo [i]->curBank, 0x2000 );
        } else {
            m_GromMemoryInfo [i] = NULL;
            memset (( USHORT * ) &m_GromMemory [ i << 13 ], 0, 0x2000 );
        }
    }

    m_Cartridge = NULL;
    if ( reset ) Reset ();
}
