//----------------------------------------------------------------------------
//
// File:	tms9919-sdl.cpp
// Date:      	20-Jun-2000
// Programmer:	Marc Rousseau
//
// Description: This file contains SDL specific code for the TMS9919
//
// Copyright (c) 2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include "common.hpp"
#include "logger.hpp"
#include "SDL.h"
#include "tms9919.hpp"
#include "tms9919-sdl.hpp"

cSdlTMS9919::cSdlTMS9919 () :
    m_Initialized ( false ),
    m_MasterVolume ( 50 ),
    m_MixBuffer ( NULL )
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919 ctor", true );

    memset ( &m_AudioSpec, 0, sizeof ( m_AudioSpec ));
    memset ( m_Info, 0, sizeof ( m_Info ));	   
		     			
    SDL_AudioSpec wanted;
    memset ( &wanted, 0, sizeof ( SDL_AudioSpec ));
			   
    // Set the audio format
    wanted.freq     = 44100;
    wanted.format   = AUDIO_S8;
    wanted.channels = 1;
    wanted.samples  = 512;
    wanted.callback = _AudioCallback;
    wanted.userdata = this;

    // Open the audio device, forcing the desired format
    if ( SDL_OpenAudio ( &wanted, &m_AudioSpec ) < 0 ) {
        ERROR ( "Couldn't open audio: " << SDL_GetError ());
    } else {  				   
        m_Initialized = true;
        m_MixBuffer = new Uint8 [ m_AudioSpec.samples ];
        memset ( m_MixBuffer, 0, sizeof ( Uint8 ) * m_AudioSpec.samples );
        SDL_PauseAudio ( false );
    }	
}

cSdlTMS9919::~cSdlTMS9919 ()
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919 dtor", true );

    if ( m_Initialized == true ) {
        SDL_PauseAudio ( true );
        SDL_CloseAudio ();
    }
    
    delete [] m_MixBuffer;
}

void cSdlTMS9919::_AudioCallback ( void *data, Uint8 *stream, int len )
{
    (( cSdlTMS9919 * ) data)->AudioCallback ( stream, len );
}

void cSdlTMS9919::AudioCallback ( Uint8 *stream, int len )
{
    int volume = ( m_MasterVolume * SDL_MIX_MAXVOLUME ) / 100;

    for ( int i = 0; i < 3; i++ ) {
        sVoiceInfo *info = &m_Info [i];
        if (( info->period == 0 ) || ( info->setting == 0 )) continue;
        for ( int j = 0; j < len; j++ ) {
            m_MixBuffer [j] = m_AudioSpec.silence + info->value;
            if ( --info->toggle == 0 ) {
                info->toggle  = info->period;
                info->setting = -info->setting;
                info->value   = info->setting;
            }
        }						  
        SDL_MixAudio ( stream, m_MixBuffer, len, volume );
    }

    // TBD: Add noise here
}

void cSdlTMS9919::SetMasterVolume ( int volume )
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919::SetMasterVolume", true );

    if ( volume == m_MasterVolume ) return;
    
    m_MasterVolume = volume;
}
    
void cSdlTMS9919::SetNoise ( int color, int type )
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919::SetNoise", true );

    if (( color == m_NoiseColor ) && ( type == m_NoiseType )) return;

    if ( m_Initialized == true ) {
    }

    cTMS9919::SetNoise ( color, type );
}

void cSdlTMS9919::SetFrequency ( int tone, int freq )
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919::SetFrequency", true );

    if ( freq == m_Frequency [ tone ] ) return;

    if ( m_Initialized == true ) {

        sVoiceInfo *info = &m_Info [tone];

        if ( freq < m_AudioSpec.freq / 2 ) {
            info->period = m_AudioSpec.freq / freq / 2;
            int newSetting = ( 128 / 3 * ( 30 - m_Attenuation [ tone ] )) / 30;
            info->setting = ( info->setting > 0 ) ? newSetting : -newSetting;
            if ( info->toggle == 0 ) {
                info->toggle = info->period;
                info->value  = info->setting;
	    }
        } else {
            info->period  = 0;
            info->toggle  = 0;
            info->setting = 0;
        }
    }

    cTMS9919::SetFrequency ( tone, freq );
}

void cSdlTMS9919::SetAttenuation ( int tone, int atten )
{
    FUNCTION_ENTRY ( this, "cSdlTMS9919::SetAttenuation", true );

    if ( atten == m_Attenuation [ tone ] ) return;

    if ( m_Initialized == true ) {
        sVoiceInfo *info = &m_Info [tone];
        int newSetting = ( 128 / 3 * ( 30 - atten )) / 30;
        info->setting = ( info->setting > 0 ) ? newSetting : -newSetting;
        if ( info->toggle == 0 ) info->value = info->setting;
    }

    cTMS9919::SetAttenuation ( tone, atten );
}
